from typing import ClassVar

from django.core.validators import RegexValidator
from django.db import models
from django.utils.functional import classproperty
from django.utils.translation import gettext_lazy as _

from localflavor.generic.models import BICField, IBANField
from payments.core import BasicProvider

from aleksis.core.mixins import ExtensiblePolymorphicModel, RegistryObject

from .base import Client


class PaymentVariant(RegistryObject, ExtensiblePolymorphicModel):
    """A single, configured payment variant."""

    icon: ClassVar[str] = None
    verbose_name: ClassVar[str] = None

    description = models.CharField(max_length=255, verbose_name=_("Description"))
    client = models.ForeignKey(
        Client, on_delete=models.CASCADE, verbose_name=_("Client"), related_name="payment_variants"
    )

    def __str__(self) -> str:
        return self.description

    def get_provider(self) -> BasicProvider:
        """Get django-payments provider for this payment variant."""
        raise NotImplementedError()

    @classproperty
    def as_choices(cls) -> list[tuple[str, str]]:
        """Get payment variants as choices."""
        return [
            (variant._class_name, variant.verbose_name) for variant in cls.registered_objects_list
        ]

    @classproperty
    def variant_name(cls):
        return cls._class_name

    class Meta:
        verbose_name = _("Payment variant")
        verbose_name_plural = _("Payment variants")


class SofortPaymentVariant(PaymentVariant):
    icon = "simple-icons:klarna"
    verbose_name = _("Klarna/Sofort")
    _class_name = "sofort"

    api_id = models.CharField(verbose_name=_("API ID"), max_length=255)
    api_key = models.CharField(verbose_name=_("API key"), max_length=255)
    project_id = models.CharField(verbose_name=_("Project ID"), max_length=255)

    def get_provider(self):
        from payments.sofort import SofortProvider  # noqa

        provider = SofortProvider(key=self.api_key, id=self.api_sid, project_id=self.project_id)
        return provider

    class Meta:
        verbose_name = _("Sofort/Klarna payment variant")
        verbose_name_plural = _("Sofort/Klarna payment variants")


class PaypalPaymentVariant(PaymentVariant):
    icon = "logos:paypal"
    verbose_name = _("PayPal")
    _class_name = "paypal"

    api_client_id = models.CharField(verbose_name=_("Client ID"), max_length=255)
    secret = models.CharField(verbose_name=_("Secret"), max_length=255)
    capture = models.BooleanField(verbose_name=_("Use PayPal Authorize & Capture"), default=False)
    endpoint = models.URLField(verbose_name=_("Endpoint"), default="https://api.paypal.com")

    def get_provider(self):
        from payments.paypal import PaypalProvider  # noqa

        provider = PaypalProvider(
            client_id=self.client_id,
            secret=self.secret,
            capture=self.capture,
            endpoint=self.endpoint,
        )
        return provider

    class Meta:
        verbose_name = _("PayPal payment variant")
        verbose_name_plural = _("PayPal payment variants")


class SEPADirectDebitPaymentVariant(PaymentVariant):
    icon = "mdi:bank-transfer"
    verbose_name = _("SEPA Direct Debit")
    _class_name = "sdd"

    creditor = models.CharField(verbose_name=_("Creditor name"), max_length=255)
    creditor_identifier = models.CharField(
        verbose_name=_("Creditor identifier"),
        max_length=35,
        validators=[RegexValidator("^[A-Z]{2}[0-9]{2}[A-Z0-9]{1,31}$")],
    )
    iban = IBANField(verbose_name=_("IBAN of bank account"))
    bic = BICField(verbose_name=_("BIC/SWIFT code of bank"))

    def get_provider(self):
        from djp_sepa.providers import DirectDebitProvider  # noqa

        provider = DirectDebitProvider(
            creditor=self.creditor,
            creditor_identifier=self.creditor_identifier,
            iban=self.iban,
            bic=self.bic,
        )
        return provider

    class Meta:
        verbose_name = _("SEPA Direct Debit payment variant")
        verbose_name_plural = _("SEPA Direct Debit payment variants")


class PledgePaymentVariant(PaymentVariant):
    icon = "mdi:hand-coin"
    verbose_name = _("Payment pledge/Manual payment")
    _class_name = "pledge"

    def get_provider(self):
        from djp_sepa.providers import PaymentPledgeProvider  # noqa

        provider = PaymentPledgeProvider()
        return provider

    class Meta:
        verbose_name = _("Pledge payment variant")
        verbose_name_plural = _("Pledge payment variants")
